/**
  * @author Horie(16th)
  * @since 2020.10.-
  * summary : 動的なtarget選択のために, Comparatorと優先度管理クラスを実装
  * 優先度と, 優先度の変化しやすさ（＝不安定度）を各対象ごとに求め管理する
  * 
  */

package Rione_2021.module.complex.taskManager;

import adf.agent.communication.MessageManager;
import adf.agent.communication.standard.bundle.MessageUtil;
import adf.agent.communication.standard.bundle.centralized.CommandPolice;
import adf.agent.communication.standard.bundle.information.MessageAmbulanceTeam;
import adf.agent.communication.standard.bundle.information.MessageFireBrigade;
import adf.agent.communication.standard.bundle.information.MessagePoliceForce;
import adf.agent.communication.standard.bundle.information.MessageRoad;
import adf.agent.develop.DevelopData;
import adf.agent.info.AgentInfo;
import adf.agent.info.ScenarioInfo;
import adf.agent.info.WorldInfo;
import adf.agent.module.ModuleManager;
import adf.agent.precompute.PrecomputeData;
import adf.component.communication.CommunicationMessage;
import adf.component.module.algorithm.PathPlanning;
import adf.component.module.algorithm.Clustering;
import adf.component.module.complex.RoadDetector;
import rescuecore2.standard.entities.*;
import rescuecore2.worldmodel.EntityID;

import java.util.*;

import static rescuecore2.standard.entities.StandardEntityURN.*;

/**
  * @author Horie(16th)
  * @since 2020.10.-
  * summary : 各基準ごとに, 優先度と優先度の変わりやすさ（不安定度）を算出する
  * interfaceであるので, これを継承して評価値計算クラスを作る 
  * 実際に使うときは, 具象クラスではなくこのクラスのインスタンスとして各インスタンスを定義する
  */

public interface RIODynamicTaskEvaluator{
    /// summary : このインスタンスが評価する対象を設定する
    public RIODynamicTaskEvaluator setEvaluatee(EntityID evaluateeID);

    /// summary : このインスタンスを利用するエージェントの情報を引き受ける
    public RIODynamicTaskEvaluator setInfo(ScenarioInfo scenarioInfo, WorldInfo worldInfo, AgentInfo agentInfo);
    public RIODynamicTaskEvaluator setAlgorithm(PathPlanning pathPlanning, Clustering clustering);

    /// summary : 各サイクルごとの事前処理
    public RIODynamicTaskEvaluator update();

    /// summary : 優先度を返す関数
    public double getPriority();

    /// summary : 不安定度を返す関数
    public double getUnstableness();
}